{-# LANGUAGE CPP, DeriveDataTypeable, PatternGuards, RelaxedPolyRec #-}
module Language.Java.Paragon.TypeCheck.Monad.CodeEnv where

import Language.Java.Paragon.Syntax
--import Language.Java.Paragon.Pretty
import Language.Java.Paragon.Interaction

--import Language.Java.Paragon.TypeCheck.Actors
import Language.Java.Paragon.TypeCheck.Policy
import Language.Java.Paragon.TypeCheck.Locks
import Language.Java.Paragon.TypeCheck.Types

import Language.Java.Paragon.TypeCheck.TypeMap

import qualified Data.Map.Strict as Map
import qualified Data.ByteString.Char8 as B

import Data.Data

codeEnvModule :: String
codeEnvModule = typeCheckerBase ++ ".Monad.CodeEnv"

data CodeEnv = CodeEnv  {
      vars      :: Map B.ByteString VarFieldSig,
      lockstate :: [TcLock],
      returnI   :: Maybe (TcType, ActorPolicy),
      exnsE     :: Map TcType (ActorPolicy, ActorPolicy),
      branchPCE :: (Map Entity [(ActorPolicy, String)], [(ActorPolicy, String)]),
      parBounds :: [(B.ByteString, ActorPolicy)]
    }
  deriving (Show, Data, Typeable)

-- Env to use when typechecking expressions not inside method
-- bodies, e.g. in field initializers and policy modifiers
simpleEnv :: ActorPolicy -> String -> CodeEnv
simpleEnv brPol str = 
    CodeEnv {
      vars = Map.empty,
      lockstate = [],
      returnI = Nothing,
      exnsE = Map.empty,
      branchPCE = (Map.empty, [(brPol,str)]),
      parBounds = []
    }

data Entity = VarEntity (Name ())
            | ThisFieldEntity B.ByteString
            | ExnEntity TcType
            | LockEntity (Name ())
            | BreakE | ContinueE | ReturnE
  deriving (Show, Eq, Ord, Data, Typeable)

varE, lockE :: Name () -> Entity
varE = VarEntity
lockE = LockEntity

exnE :: TcType -> Entity
exnE = ExnEntity

thisFE :: B.ByteString -> Entity
thisFE = ThisFieldEntity 

breakE, continueE, returnE :: Entity
breakE = BreakE
continueE = ContinueE
returnE = ReturnE

--------------------------------------
--    Working with the branchPC     --
--------------------------------------

branchPC :: Maybe Entity -> CodeEnv -> [(ActorPolicy, String)]
branchPC men (CodeEnv { branchPCE = (bm, def) }) = 
    flip (maybe def) men $ \en ->
        maybe def id (Map.lookup en bm)

joinBranchPC :: ActorPolicy -> String -> CodeEnv -> CodeEnv
joinBranchPC p str env = let (bm, def) = branchPCE env
                         in env { branchPCE = (Map.map ((p, str):) bm, (p,str):def) }
